/*
	AIL Ruriiro no YUki
	for LSI-C(86)
*/
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>


unsigned long decodeLZ_0( FILE *in, unsigned long size, FILE *out )
{
	unsigned char	buf[256];
	unsigned char	*dic;
	unsigned long	dicpos = 0x0EFE;
	unsigned char	byte;
	unsigned char	sign;
	unsigned int	decode_startaddr;
	unsigned int	decode_length;
	int				i;
	int				q;
	unsigned long	decode_byte = 0L;
	unsigned long	read_byte = 0L;
	unsigned long	addr = 0L;

	dic = ( unsigned char * )calloc( 1, 4096 + 10 );

	while( read_byte < size )
	{
		/* compress sign */
		fread( buf, 1, 1, in ); sign = buf[0]; sign ^= 0xFF;
		read_byte++;

		if( read_byte >= size ) break;

		for( i = 0; i != 8; i++ )
		{
			if( sign & 1 )
			{
				fread( buf, 1, 1, in ); byte = buf[0];
				dic[dicpos] = byte;
				fprintf( out, "%c", byte );
				dicpos++; dicpos &= 0x0FFF;
				read_byte++; decode_byte++;
			} else {
				fread( buf, 3, 1, in );
				decode_length = buf[2] + 3;
				decode_startaddr  = (( buf[1] & 0x0F ) << 8 );
				decode_startaddr += buf[0];
				read_byte += 3;

				/* decode */
				for( q = 0; q != decode_length; q++ )
				{
					addr = decode_startaddr + q; addr &= 0x0FFF;
					dic[dicpos] = dic[addr];
					fprintf( out, "%c", dic[addr] );
					dicpos++; dicpos &= 0x0FFF;
					decode_byte++;
				}
			}
			sign >>= 1;
			if( read_byte >= size ) break;
		}
	}
	free( dic );
	return decode_byte;
}


unsigned long decodeLZ( FILE *in, unsigned long size, FILE *out )
{
	unsigned char	buf[256];
	unsigned char	*dic;
	unsigned long	dicpos = 0x0FEE;
	unsigned char	byte;
	unsigned char	sign;
	unsigned int	decode_startaddr;
	unsigned int	decode_length;
	int				i;
	int				q;
	unsigned long	decode_byte = 0L;
	unsigned long	read_byte = 0L;
	unsigned long	addr = 0L;

	dic = ( unsigned char * )calloc( 1, 4096 + 10 );

	while( read_byte < size )
	{
		/* compress sign */
		fread( buf, 1, 1, in ); sign = buf[0]; sign ^= 0xFF;
		read_byte++;

		if( read_byte >= size ) break;

		for( i = 0; i != 8; i++ )
		{
			if( sign & 1 )
			{
				fread( buf, 1, 1, in ); byte = buf[0];
				dic[dicpos] = byte;
				fprintf( out, "%c", byte );
				dicpos++; dicpos &= 0x0FFF;
				read_byte++; decode_byte++;
			} else {
				fread( buf, 2, 1, in );
				decode_length = ( buf[1] & 0x0F ) + 3;
				decode_startaddr  = (( buf[1] & 0xF0 ) << 4 );
				decode_startaddr += buf[0];
				read_byte += 2;

				/* decode */
				for( q = 0; q != decode_length; q++ )
				{
					addr = decode_startaddr + q; addr &= 0x0FFF;
					dic[dicpos] = dic[addr];
					fprintf( out, "%c", dic[addr] );
					dicpos++; dicpos &= 0x0FFF;
					decode_byte++;
				}
			}
			sign >>= 1;
			if( read_byte >= size ) break;
		}
	}
	free( dic );
	return decode_byte;
}

void main( int argc, char **argv )
{
	FILE			*in;
	FILE			*out;
	unsigned long	filesize_compress = 0L;
	unsigned long	filesize_original = 0L;
	char			out_filename[256];
	char			buf[2];

	printf( "AIL Ruriiro no Yuki Music Decompress tool 070917 (c)Fac.KUROHANE\n" );

	if( argc == 1 )
	{
		printf( "usage: RURIEXT <filename> <dest_dir>\n" );
		return;
	}

	in = fopen( argv[1], "rb" );
	if( in == NULL )
	{
		printf( "error: can't open %s\n", argv[1] );
		return;
	}
	printf( "filename: %s\n", argv[1] );

	/* filesize */
	fseek( in, 0, SEEK_END );
	filesize_compress = ftell( in );
	fseek( in, 0, SEEK_SET );

	printf( "fileisize_compress: %lu(%04lX)\n",
		filesize_compress,
		filesize_compress
	);

	if( argc == 3 )
	{
		sprintf( out_filename, "%s%s",
			argv[2],
			argv[1]
		);
	} else sprintf( out_filename, "%s", argv[1] );

	out = fopen( out_filename, "wb+" );
	if( out == NULL )
	{
		printf( "error: can't write %s.\n", out_filename );
		goto end;
	}

	filesize_compress -= 2;

	/* LZ extract */
	fread( buf, 1, 2, in );
	if( buf[1] != 0 )
	{
		fclose( out );
		printf( "error: Unknown type %s.\n", argv[1] );
		goto end;
	}

	if( buf[0] == 0 )
	{
		filesize_original = decodeLZ_0( in, filesize_compress, out );
		printf( "filesize_original: %lu(%04lX)\n",
			filesize_original,
			filesize_original
		);
	} else if( buf[0] == 1 )
	{
		filesize_original = decodeLZ( in, filesize_compress, out );
		printf( "filesize_original: %lu(%04lX)\n",
			filesize_original,
			filesize_original
		);
	} else {
		fclose( out );
		printf( "error: Unknown type %s.\n", argv[1] );
		goto end;
	}

	fclose( in );
	fclose( out );

	printf( "\n" );
	printf( "complete.\n" );

	return;

end:;
	fclose( in );
}
